# -*- coding: utf-8 -*-
"""
    Market quote and trade context setting
"""
from futu import *
from futu.common.conn_mng import *
from futu.common.utils import *

class Request(object):

    def __init__(self):
        self.begin = 0  # 数据起始点
        self.num = 200  # 请求数据个数，最大200
        self.sort_field = SortField.CODE  # 根据哪个字段排序
        self.ascend = True  # 升序ture, 降序false
        self.stock_owner = None  # 所属正股
        self.type_list = list()   # Qot_Common.WarrantType, 窝轮类型过滤列表 WrtType
        self.issuer_list = list()  # Qot_Common.Issuer, 发行人过滤列表
        self.maturity_time_min = None # 到期日, 到期日范围的开始时间戳
        self.maturity_time_max = None # 到期日范围的结束时间戳
        self.ipo_period = None  # 上市日
        self.price_type = None  # Qot_Common.PriceType, 价内 / 价外
        self.status = None  # Qot_Common.WarrantStatus, 窝轮状态
        self.cur_price_min = None  # 最新价过滤起点
        self.cur_price_max = None  # 最新价过滤终点
        self.strike_price_min = None  # 行使价过滤起点
        self.strike_price_max = None  # 行使价过滤终点
        self.street_min = None  # 街货占比 % 过滤起点
        self.street_max = None  # 街货占比 % 过滤终点
        self.conversion_min = None  # 换股比率过滤起点
        self.conversion_max = None  # 换股比率过滤终点
        self.vol_min = None  # 成交量过滤起点
        self.vol_max = None  # 成交量过滤终点
        self.premium_min = None  # 溢价 % 过滤起点
        self.premium_max = None  # 溢价 % 过滤终点
        self.leverage_ratio_min = None  # 杠杆比率过滤起点
        self.leverage_ratio_max = None  # 杠杆比率过滤终点
        self.delta_min = None   # 对冲值过滤起点, 仅认购认沽支持该字段过滤
        self.delta_max = None  # 对冲值过滤终点, 仅认购认沽支持该字段过滤
        self.implied_min = None  # 引伸波幅过滤起点, 仅认购认沽支持该字段过滤
        self.implied_max = None  # 引伸波幅过滤终点, 仅认购认沽支持该字段过滤
        self.recovery_price_min = None  # 回收价过滤起点, 仅牛熊证支持该字段过滤
        self.recovery_price_max = None  # 回收价过滤终点, 仅牛熊证支持该字段过滤
        self.price_recovery_ratio_min = None  # 正股距回收价 % 过滤起点, 仅牛熊证支持该字段过滤
        self.price_recovery_ratio_max = None  # 正股距回收价 % 过滤终点, 仅牛熊证支持该字段过滤

    def fill_request_pb(self):
        from futu.common.pb.Qot_GetWarrant_pb2 import Request as GetWarrantPBRequest
        pb = GetWarrantPBRequest()
        pb.c2s.begin = self.begin
        pb.c2s.num = self.num

        r, v = SortField.to_number(self.sort_field)
        if not r:
            return RET_ERROR, 'sort_field is wrong. must be SortField'
        pb.c2s.sortField = v

        if type(self.ascend) != bool:
            return RET_ERROR, 'ascend is wrong. must be bool'

        pb.c2s.ascend = self.ascend
        if self.stock_owner is not None and len(self.stock_owner) != 0:
            """所属正股"""
            ret, content = split_stock_str(self.stock_owner)
            if ret == RET_ERROR:
                error_str = content
                return RET_ERROR, error_str
            market_code, stock_code = content
            pb.c2s.owner.market = market_code
            pb.c2s.owner.code = stock_code

        if self.type_list is not None and not isinstance(self.type_list, list):
            # 1120400921001028867
            self.type_list = [self.type_list]
        if self.issuer_list is not None and not isinstance(self.issuer_list, list):
            # 1120400921001028867
            self.issuer_list = [self.issuer_list]

        if self.type_list is not None and len(self.type_list) != 0:
            """Qot_Common.WarrantType,窝轮类型过滤列表"""
            for t in self.type_list:
                r, v = WrtType.to_number(t)
                if not r:
                    return RET_ERROR, 'type_list is wrong. must be [WrtType]'
                pb.c2s.typeList.append(v)
        if self.issuer_list is not None and len(self.issuer_list) != 0:
            """Qot_Common.Issuer,发行人过滤列表"""
            for t in self.issuer_list:
                r, v = Issuer.to_number(t)
                if not r:
                    return RET_ERROR, 'issuer_list is wrong. must be [Issuer]'
                pb.c2s.issuerList.append(v)

        if self.maturity_time_min is not None:
            """到期日, 到期日范围的开始时间戳"""
            if not isinstance(self.maturity_time_min, str) or len(self.maturity_time_min) == 0:
                return RET_ERROR, 'maturity_time_min is wrong. must be str. eg:2018-02-05'
            pb.c2s.maturityTimeMin = self.maturity_time_min

        if self.maturity_time_max is not None and len(self.maturity_time_max) != 0:
            """到期日范围的结束时间戳"""
            if not isinstance(self.maturity_time_max, str) or len(self.maturity_time_max) == 0:
                return RET_ERROR, 'maturity_time_max is wrong. must be str. eg:2018-02-10'
            pb.c2s.maturityTimeMax = self.maturity_time_max

        if self.ipo_period is not None:
            """上市日"""
            r, v = IpoPeriod.to_number(self.ipo_period)
            if not r:
                return RET_ERROR, 'ipo_period is wrong. must be IpoPeriod'
            pb.c2s.ipoPeriod = v
        if self.price_type is not None:
            """价内/价外"""
            r, v = PriceType.to_number(self.price_type)
            if not r:
                return RET_ERROR, 'price_type is wrong. must be PriceType'
            pb.c2s.priceType = v
        if self.status is not None:
            """窝轮状态"""
            r, v = WarrantStatus.to_number(self.status)
            if not r:
                return RET_ERROR, 'status is wrong. must be WarrantStatus'
            pb.c2s.status = v
        if self.cur_price_min is not None:
            """最新价过滤起点"""
            pb.c2s.curPriceMin = self.cur_price_min
        if self.cur_price_max is not None:
            """最新价过滤终点"""
            pb.c2s.curPriceMax = self.cur_price_max
        if self.strike_price_min is not None:
            """行使价过滤起点"""
            pb.c2s.strikePriceMin = self.strike_price_min
        if self.strike_price_max is not None:
            """行使价过滤终点 """
            pb.c2s.strikePriceMax = self.strike_price_max
        if self.street_min is not None:
            """街货占比%过滤起点"""
            pb.c2s.streetMin = self.street_min
        if self.street_max is not None:
            """街货占比%过滤终点"""
            pb.c2s.streetMax = self.street_max
        if self.conversion_min is not None:
            """换股比率过滤起点"""
            pb.c2s.conversionMin = self.conversion_min
        if self.conversion_max is not None:
            """换股比率过滤终点"""
            pb.c2s.conversionMax = self.conversion_max
        if self.vol_min is not None and self.vol_min != -1:
            """成交量过滤起点"""
            pb.c2s.volMin = self.vol_min
        if self.vol_max is not None and self.vol_max != -1:
            """成交量过滤终点"""
            pb.c2s.volMax = self.vol_max
        if self.premium_min is not None:
            """溢价 % 过滤起点"""
            pb.c2s.premiumMin = self.premium_min
        if self.premium_max is not None:
            """溢价 % 过滤终点"""
            pb.c2s.premiumMax = self.premium_max
        if self.leverage_ratio_min is not None:
            """杠杆比率过滤起点"""
            pb.c2s.leverageRatioMin = self.leverage_ratio_min
        if self.leverage_ratio_max is not None:
            """杠杆比率过滤终点"""
            pb.c2s.leverageRatioMax = self.leverage_ratio_max
        if self.delta_min is not None:
            """对冲值过滤起点, 仅认购认沽支持该字段过滤"""
            pb.c2s.deltaMin = self.delta_min
        if self.delta_max is not None:
            """对冲值过滤起点, 仅认购认沽支持该字段过滤"""
            pb.c2s.deltaMax = self.delta_max
        if self.implied_min is not None:
            """引伸波幅过滤起点,仅认购认沽支持该字段过滤"""
            pb.c2s.impliedMin = self.implied_min
        if self.implied_max is not None:
            """引伸波幅过滤终点,仅认购认沽支持该字段过滤"""
            pb.c2s.impliedMax = self.implied_max
        if self.recovery_price_min is not None:
            """回收价过滤起点,仅牛熊证支持该字段过滤"""
            pb.c2s.recoveryPriceMin = self.recovery_price_min
        if self.recovery_price_max is not None:
            """回收价过滤终点,仅牛熊证支持该字段过滤"""
            pb.c2s.recoveryPriceMax = self.recovery_price_max
        if self.price_recovery_ratio_min is not None:
            """正股距回收价%过滤起点,仅牛熊证支持该字段过滤"""
            pb.c2s.priceRecoveryRatioMin = self.price_recovery_ratio_min
        if self.price_recovery_ratio_max is not None:
            """正股距回收价%过滤终点,仅牛熊证支持该字段过滤"""
            pb.c2s.priceRecoveryRatioMax = self.price_recovery_ratio_max

        return RET_OK, pb


class Response(object):

    @staticmethod
    def unpack_response_pb(resp):
        from futu.common.pb.Qot_GetWarrant_pb2 import Response as GetWarrantPBResponse
        if resp is None or not isinstance(resp, GetWarrantPBResponse):
            return RET_ERROR, "unpack_response_pb error", None
        if resp.retType != RET_OK:
            return RET_ERROR, resp.retMsg, None

        warrant_data_list = list()

        for item in resp.s2c.warrantDataList:
            warrant = dict()
            warrant["stock"] = merge_qot_mkt_stock_str(
                int(item.stock.market), item.stock.code)  # 股票
            warrant["stock_owner"] = merge_qot_mkt_stock_str(
                int(item.owner.market), item.owner.code)  # 所属正股
            warrant["type"] = WrtType.to_string2(item.type)  # 窝轮类型
            warrant["issuer"] = Issuer.to_string2(item.issuer)  # 发行人
            warrant["maturity_time"] = item.maturityTime  # 到期日
            if item.HasField("maturityTimestamp"):
                """到期日时间戳"""
                warrant["maturity_timestamp"] = item.maturityTimestamp
            if item.HasField("listTime"):
                """上市时间"""
                warrant["list_time"] = item.listTime
            if item.HasField("listTimestamp"):
                """上市时间戳"""
                warrant["list_timestamp"] = item.listTimestamp
            if item.HasField("lastTradeTime"):
                """最后交易日"""
                warrant["last_trade_time"] = item.lastTradeTime
            if item.HasField("lastTradeTimestamp"):
                """最后交易日时间戳"""
                warrant["last_trade_timestamp"] = item.lastTradeTimestamp
            if item.HasField("recoveryPrice"):
                """回收价,仅牛熊证支持该字段"""
                warrant["recovery_price"] = item.recoveryPrice
            if item.HasField("conversionRatio"):
                """换股比率"""
                warrant["conversion_ratio"] = item.conversionRatio
            if item.HasField("lotSize"):
                """每手数量"""
                warrant["lot_size"] = item.lotSize
            if item.HasField("strikePrice"):
                """行使价"""
                warrant["strike_price"] = item.strikePrice
            if item.HasField("lastClosePrice"):
                """昨收价"""
                warrant["last_close_price"] = item.lastClosePrice
            if item.HasField("curPrice"):
                """当前价"""
                warrant["cur_price"] = item.curPrice
            if item.HasField("priceChangeVal"):
                """涨跌额"""
                warrant["price_change_val"] = item.priceChangeVal
            if item.HasField("changeRate"):
                """涨跌幅%"""
                warrant["change_rate"] = item.changeRate
            if item.HasField("status"):
                """窝轮状态"""
                warrant["status"] = WarrantStatus.to_string2(item.status)
            if item.HasField("bidPrice"):
                """买入价"""
                warrant["bid_price"] = item.bidPrice
            if item.HasField("askPrice"):
                """卖出价"""
                warrant["ask_price"] = item.askPrice
            if item.HasField("bidVol"):
                """买量"""
                warrant["bid_vol"] = item.bidVol
            if item.HasField("askVol"):
                """卖量"""
                warrant["ask_vol"] = item.askVol
            if item.HasField("volume"):
                """成交量"""
                warrant["volume"] = item.volume
            if item.HasField("turnover"):
                """成交额"""
                warrant["turnover"] = item.turnover
            if item.HasField("score"):
                """综合评分"""
                warrant["score"] = item.score
            if item.HasField("premium"):
                """溢价%"""
                warrant["premium"] = item.premium
            if item.HasField("breakEvenPoint"):
                """打和点"""
                warrant["break_even_point"] = item.breakEvenPoint
            if item.HasField("leverage"):
                """杠杠比率"""
                warrant["leverage"] = item.leverage
            if item.HasField("ipop"):
                """价内/价外%"""
                warrant["ipop"] = item.ipop
            if item.HasField("priceRecoveryRatio"):
                """正股距回收价%，仅牛熊证支持该字段"""
                warrant["price_recovery_ratio"] = item.priceRecoveryRatio
            if item.HasField("conversionPrice"):
                """换股价"""
                warrant["conversion_price"] = item.conversionPrice
            if item.HasField("streetRate"):
                """街货占比"""
                warrant["street_rate"] = item.streetRate
            if item.HasField("streetVol"):
                """街货量"""
                warrant["street_vol"] = item.streetVol
            if item.HasField("amplitude"):
                """振幅%"""
                warrant["amplitude"] = item.amplitude
            if item.HasField("name"):
                """name"""
                warrant["name"] = item.name
            if item.HasField("issueSize"):
                """发行量"""
                warrant["issue_size"] = item.issueSize
            if item.HasField("highPrice"):
                """最高价"""
                warrant["high_price"] = item.highPrice
            if item.HasField("lowPrice"):
                """最低价"""
                warrant["low_price"] = item.lowPrice
            if item.HasField("impliedVolatility"):
                """引申波幅,仅认购认沽支持该字段"""
                warrant["implied_volatility"] = item.impliedVolatility
            if item.HasField("delta"):
                """对冲值,仅认购认沽支持该字段"""
                warrant["delta"] = item.delta
            if item.HasField("effectiveLeverage"):
                """有效杠杆"""
                warrant["effective_leverage"] = item.effectiveLeverage
            if item.HasField("upperStrikePrice"):
                """上限价，仅界内证支持该字段"""
                warrant["upper_strike_price"] = item.upperStrikePrice
            if item.HasField("lowerStrikePrice"):
                """下限价，仅界内证支持该字段"""
                warrant["lower_strike_price"] = item.lowerStrikePrice
            if item.HasField("inLinePriceStatus"):
                """界内界外，仅界内证支持该字段"""
                warrant["inline_price_status"] = PriceType.to_string2(
                    item.inLinePriceStatus)
            warrant_data_list.append(warrant)
        return RET_OK, "", (warrant_data_list, resp.s2c.lastPage, resp.s2c.allCount)


if __name__ == "__main__":
    req = Request()
    req.type_list.append(WrtType.BUY)
    req.fill_request_pb()



# -*- coding: utf-8 -*-

"""演示如何使用股票筛选功能"""

from futu import *

quote_ctx = OpenQuoteContext(host='127.0.0.1', port=11111)

field = SimpleFilter()
field.filter_min = 100
field.filter_max = 1000
field.stock_field = StockField.CUR_PRICE
field.is_no_filter = False
field.sort = SortDir.ASCEND

# 可选择增加更多筛选条件
# field2 = SimpleFilter()
# field2.filter_min = 100
# field2.filter_max = 1000
# field2.stock_field = StockField.VOLUME_RATIO
# field2.sort = SortDir.ASCEND
# field2.is_no_filter = False
#
# field3 = SimpleFilter()
# field3.stock_field = StockField.CUR_PRICE_TO_HIGHEST52_WEEKS_RATIO
# field3.is_no_filter = True

ret, ls = quote_ctx.get_stock_filter(Market.HK, [field])
last_page, all_count, ret_list = ls
print(ret_list)

quote_ctx.close()





import pandas as pd
from futu.quote.quote_query import *


class OpenQuoteTool(object):

    def __init__(self, context):
        self.quote_context = context
        if context is None:
            raise Exception("context is empty!")

    def save_delay_statistics(self, file_path):
        if file_path is None:
            return
        html_file = open(file_path, 'w', encoding='utf-8')
        local_path = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(local_path, "head.html"), 'r', encoding='utf-8') as html_head_file:
            html_file.write(html_head_file.read())

        segment_list = [0, 2, 4, 6, 8, 10, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56,
                        60, 64, 68, 72, 76, 80, 84, 88, 92, 96, 100, 110, 120, 130, 140,
                        150, 160, 170, 180, 200, 250, 300, 500, 1000, -1]
        req_head_describe = [("proto_id", "命令字"),
                             ("count", "请求个数"),
                             ("is_local_reply", "是否本地数据"),
                             ("total_cost_avg", "总体延迟平均值（ms）")]

        push_head_describe = [("begin", "速度区间ms（开始）"),
                              ("end", "速度区间ms（结束）"),
                              ("proportion", "当前区段占比"),
                              ("count", "包数"),
                              ("cumulative_ratio", "累计占比")]

        for stage_type in QotPushStage.ALL[::-1]:

            if stage_type == QotPushStage.SR2_CS:
                delay_statistics_type = DelayStatisticsType.ALL
            else:
                delay_statistics_type = [DelayStatisticsType.QOT_PUSH]
            err, dic = self.quote_context.get_delay_statistics(
                delay_statistics_type,
                stage_type,
                segment_list)
            if err != RET_OK:
                return

            qot_pushes = dic["qot_push"]
            req_replies = dic["req_reply"]
            place_orders = dic["place_order"]
            print(qot_pushes, req_replies, place_orders)

            html_file.write(
                "<div class='title'>{}</div><HR>\n".format(QotPushStage.get_describe(stage_type)))

            if len(qot_pushes) > 0:
                html_file.write("<div class='sub-title'>推送耗时统计</div>\n")

                col_list = list()
                col_new_head = list()
                for t in push_head_describe:
                    col_list.append(t[0])
                    col_new_head.append(t[1])

                for push in qot_pushes:
                    df = pd.DataFrame(push["list"], columns=col_list)
                    df.drop(df[df.proportion == 0.0].index, inplace=True)
                    if len(df) == 0:
                        continue
                    df['proportion'] = df['proportion'].map(
                        lambda x: format(x/100, '.2%'))
                    df['cumulative_ratio'] = df['cumulative_ratio'].map(
                        lambda x: format(x / 100, '.2%'))

                    # df["cumulative_ratio"] = df["begin"].map(str) + ' - ' + df["end"].map(str)
                    df.insert(0, 'interval', '')
                    try:
                        df["interval"] = df.apply(lambda x: str(
                            x.begin) + 'ms - ' + str(x.end) + "ms", axis=1)
                    except Exception as e:
                        print("df.apply error", e)
                        pass

                    df.drop(['begin', 'end'], axis=1,
                            inplace=True)  # 删除begin、end
                    col_new_head = ['统计区间（ms）', '百分比', '包数', '总体延迟百分比']
                    df.columns = col_new_head
                    qot_push_type = QotPushType.to_string2(
                        push["qot_push_type"])
                    html_file.write(
                        "<div class='item'>{}推送耗时</div>".format(QotPushType.get_describe(qot_push_type)))
                    html_file.write(df.to_html(header=True, index=False))
                    html_file.write("<br>\n")

            if len(req_replies) > 0:
                html_file.write("<div class='sub-title'>命令字请求耗时统计</div>")
                col_list = list()
                col_new_head = list()
                for t in req_head_describe:
                    col_list.append(t[0])
                    col_new_head.append(t[1])
                df = pd.DataFrame(req_replies, columns=col_list)
                df.columns = col_new_head
                html_file.write(df.to_html(header=True, index=False))
                html_file.write("<br>\n")

            if len(place_orders) > 0:
                html_file.write("<div class='sub-title'>下单统计耗时</div>")
                df = pd.DataFrame(place_orders)
                html_file.write(df.to_html(header=True, index=False))
                html_file.write("<br><br>\r\n")

        html_file.write("</body></html>")
        html_file.close()



# -*- coding: utf-8 -*-
import pandas as pd
from futu.common import RspHandlerBase
from futu.trade.trade_query import *


class TradeOrderHandlerBase(RspHandlerBase):
    """sys notify"""
    def on_recv_rsp(self, rsp_pb):
        """receive response callback function"""
        ret_code, ret_data = UpdateOrderPush.unpack_rsp(rsp_pb)

        if ret_code != RET_OK:
            return ret_code, ret_data
        else:
            order_dict = ret_data
            col_list = ['trd_env', 'code', 'stock_name', 'dealt_avg_price', 'dealt_qty',
                        'qty', 'order_id', 'order_type', 'price', 'order_status',
                        'create_time', 'updated_time', 'trd_side', 'last_err_msg', 'trd_market', "remark",
                        "time_in_force", "fill_outside_rth"
                        ]

            trade_frame_table = pd.DataFrame([order_dict], columns=col_list)
            return RET_OK, trade_frame_table


class TradeDealHandlerBase(RspHandlerBase):
    """sys notify"""
    def on_recv_rsp(self, rsp_pb):
        """receive response callback function"""
        ret_code, ret_data = UpdateDealPush.unpack_rsp(rsp_pb)

        if ret_code != RET_OK:
            return ret_code, ret_data
        else:
            deal_dict = ret_data
            col_list = ['trd_env', 'code', 'stock_name', 'deal_id', 'order_id',
                        'qty', 'price', 'trd_side', 'create_time', 'counter_broker_id',
                        'counter_broker_name', 'trd_market', 'status'
                        ]

            trade_frame_table = pd.DataFrame([deal_dict], columns=col_list)
            return RET_OK, trade_frame_table





# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Trd_ModifyOrder.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import Common_pb2 as Common__pb2
import Trd_Common_pb2 as Trd__Common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='Trd_ModifyOrder.proto',
  package='Trd_ModifyOrder',
  syntax='proto2',
  serialized_pb=_b('\n\x15Trd_ModifyOrder.proto\x12\x0fTrd_ModifyOrder\x1a\x0c\x43ommon.proto\x1a\x10Trd_Common.proto\"\xd5\x01\n\x03\x43\x32S\x12\"\n\x08packetID\x18\x01 \x02(\x0b\x32\x10.Common.PacketID\x12%\n\x06header\x18\x02 \x02(\x0b\x32\x15.Trd_Common.TrdHeader\x12\x0f\n\x07orderID\x18\x03 \x02(\x04\x12\x15\n\rmodifyOrderOp\x18\x04 \x02(\x05\x12\x0e\n\x06\x66orAll\x18\x05 \x01(\x08\x12\x0b\n\x03qty\x18\x08 \x01(\x01\x12\r\n\x05price\x18\t \x01(\x01\x12\x13\n\x0b\x61\x64justPrice\x18\n \x01(\x08\x12\x1a\n\x12\x61\x64justSideAndLimit\x18\x0b \x01(\x01\"=\n\x03S2C\x12%\n\x06header\x18\x01 \x02(\x0b\x32\x15.Trd_Common.TrdHeader\x12\x0f\n\x07orderID\x18\x02 \x02(\x04\",\n\x07Request\x12!\n\x03\x63\x32s\x18\x01 \x02(\x0b\x32\x14.Trd_ModifyOrder.C2S\"e\n\x08Response\x12\x15\n\x07retType\x18\x01 \x02(\x05:\x04-400\x12\x0e\n\x06retMsg\x18\x02 \x01(\t\x12\x0f\n\x07\x65rrCode\x18\x03 \x01(\x05\x12!\n\x03s2c\x18\x04 \x01(\x0b\x32\x14.Trd_ModifyOrder.S2CBE\n\x13\x63om.futu.openapi.pbZ.github.com/futuopen/ftapi4go/pb/trdmodifyorder')
  ,
  dependencies=[Common__pb2.DESCRIPTOR,Trd__Common__pb2.DESCRIPTOR,])




_C2S = _descriptor.Descriptor(
  name='C2S',
  full_name='Trd_ModifyOrder.C2S',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='packetID', full_name='Trd_ModifyOrder.C2S.packetID', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='header', full_name='Trd_ModifyOrder.C2S.header', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='orderID', full_name='Trd_ModifyOrder.C2S.orderID', index=2,
      number=3, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='modifyOrderOp', full_name='Trd_ModifyOrder.C2S.modifyOrderOp', index=3,
      number=4, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='forAll', full_name='Trd_ModifyOrder.C2S.forAll', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='qty', full_name='Trd_ModifyOrder.C2S.qty', index=5,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='price', full_name='Trd_ModifyOrder.C2S.price', index=6,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adjustPrice', full_name='Trd_ModifyOrder.C2S.adjustPrice', index=7,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adjustSideAndLimit', full_name='Trd_ModifyOrder.C2S.adjustSideAndLimit', index=8,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=288,
)


_S2C = _descriptor.Descriptor(
  name='S2C',
  full_name='Trd_ModifyOrder.S2C',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='Trd_ModifyOrder.S2C.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='orderID', full_name='Trd_ModifyOrder.S2C.orderID', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=290,
  serialized_end=351,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='Trd_ModifyOrder.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='c2s', full_name='Trd_ModifyOrder.Request.c2s', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=353,
  serialized_end=397,
)


_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='Trd_ModifyOrder.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='retType', full_name='Trd_ModifyOrder.Response.retType', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=-400,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='retMsg', full_name='Trd_ModifyOrder.Response.retMsg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='errCode', full_name='Trd_ModifyOrder.Response.errCode', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='s2c', full_name='Trd_ModifyOrder.Response.s2c', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=399,
  serialized_end=500,
)

_C2S.fields_by_name['packetID'].message_type = Common__pb2._PACKETID
_C2S.fields_by_name['header'].message_type = Trd__Common__pb2._TRDHEADER
_S2C.fields_by_name['header'].message_type = Trd__Common__pb2._TRDHEADER
_REQUEST.fields_by_name['c2s'].message_type = _C2S
_RESPONSE.fields_by_name['s2c'].message_type = _S2C
DESCRIPTOR.message_types_by_name['C2S'] = _C2S
DESCRIPTOR.message_types_by_name['S2C'] = _S2C
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

C2S = _reflection.GeneratedProtocolMessageType('C2S', (_message.Message,), dict(
  DESCRIPTOR = _C2S,
  __module__ = 'Trd_ModifyOrder_pb2'
  # @@protoc_insertion_point(class_scope:Trd_ModifyOrder.C2S)
  ))
_sym_db.RegisterMessage(C2S)

S2C = _reflection.GeneratedProtocolMessageType('S2C', (_message.Message,), dict(
  DESCRIPTOR = _S2C,
  __module__ = 'Trd_ModifyOrder_pb2'
  # @@protoc_insertion_point(class_scope:Trd_ModifyOrder.S2C)
  ))
_sym_db.RegisterMessage(S2C)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), dict(
  DESCRIPTOR = _REQUEST,
  __module__ = 'Trd_ModifyOrder_pb2'
  # @@protoc_insertion_point(class_scope:Trd_ModifyOrder.Request)
  ))
_sym_db.RegisterMessage(Request)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), dict(
  DESCRIPTOR = _RESPONSE,
  __module__ = 'Trd_ModifyOrder_pb2'
  # @@protoc_insertion_point(class_scope:Trd_ModifyOrder.Response)
  ))
_sym_db.RegisterMessage(Response)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\023com.futu.openapi.pbZ.github.com/futuopen/ftapi4go/pb/trdmodifyorder'))
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
"""
Examples for use the python functions: get push data
"""
from time import sleep
from futu import *

#打印数据不全请把以下注释打开
#import pandas as pd
#pd.set_option('display.height', 10000)
#pd.set_option('display.max_rows', 500)
#pd.set_option('display.max_columns', 500)
#pd.set_option('display.width', 1000)

class StockQuoteTest(StockQuoteHandlerBase):
    """
    获得报价推送数据
    """
    def on_recv_rsp(self, rsp_pb):
        """数据响应回调函数"""
        ret_code, content = super(StockQuoteTest, self).on_recv_rsp(rsp_pb)
        if ret_code != RET_OK:
            logger.debug("StockQuoteTest: error, msg: %s" % content)
            return RET_ERROR, content
        #需要打印数据把以下注释打开，其他回调数据同样处理即可
        #else:
        #    print(content)
        return RET_OK, content


class TickerTest(TickerHandlerBase):
    """ 获取逐笔推送数据 """
    def on_recv_rsp(self, rsp_pb):
        """数据响应回调函数"""
        ret_code, content = super(TickerTest, self).on_recv_rsp(rsp_pb)
        if ret_code != RET_OK:
            print("* TickerTest: error, msg: %s" % content)
            return RET_ERROR, content
        return RET_OK, content


class OrderBookTest(OrderBookHandlerBase):
    """ 获得摆盘推送数据 """
    def on_recv_rsp(self, rsp_pb):
        """数据响应回调函数"""
        ret_code, content = super(OrderBookTest, self).on_recv_rsp(rsp_pb)
        if ret_code != RET_OK:
            print("* OrderBookTest: error, msg: %s" % content)
            return RET_ERROR, content
        return RET_OK, content


class BrokerTest(BrokerHandlerBase):
    """ 获取经纪队列推送数据 """
    def on_recv_rsp(self, rsp_pb):
        """数据响应回调函数"""
        ret_code, stock_code, contents = super(BrokerTest, self).on_recv_rsp(rsp_pb)
        if ret_code != RET_OK:
            print("* BrokerTest: error, msg: %s" % contents)
            return RET_ERROR, contents
        return ret_code


def quote_test():
    '''
    行情接口调用测试
    :return:
    '''
    quote_ctx = OpenQuoteContext(host='127.0.0.1', port=11111)

    # 设置异步回调接口
    quote_ctx.set_handler(StockQuoteTest())
    quote_ctx.set_handler(TickerTest())
    quote_ctx.set_handler(OrderBookTest())
    quote_ctx.set_handler(BrokerTest())
    quote_ctx.start()

    # 获取推送数据
    subtype_list = [SubType.QUOTE, SubType.ORDER_BOOK, SubType.TICKER, SubType.BROKER]
    sub_codes = ['HK.00700', 'HK.999010']
    print("* subscribe : {}\n".format(quote_ctx.subscribe(sub_codes, subtype_list)))


if __name__ =="__main__":
    set_futu_debug_model(True)
    ''' 行情api测试 '''
    quote_test()






import sys
import os
import os.path
import platform
import futu
import multiprocessing as mp

if futu.IS_PY2:
    import Queue as queue
else:
    import queue


def print_sys_info(opend_ip=None, opend_port=None):
    if futu.IS_PY2:
        mp.freeze_support()
    opend_version = get_opend_version(opend_ip, opend_port)
    futu_path = os.path.abspath(os.path.realpath(futu.__file__))
    log_dir = _get_log_dir()

    print('Futu path: ',  futu_path)
    print('Futu version: ', futu.__version__)
    print('OpenD version:', opend_version)
    print('Python path: ', sys.executable)
    print('Python version: ', platform.python_version())
    print('OS: ', sys.platform)
    print('Platform: ', platform.platform())
    print('Arch: ', platform.architecture())
    print('Module search path: ', sys.path)
    print('Log dir: ', log_dir)


def get_opend_version(opend_ip, opend_port):
    q = mp.Queue()
    proc = mp.Process(target=_opend_proc, args=(q, opend_ip, opend_port))
    try:
        proc.start()
        opend_version = q.get(timeout=5)
        return opend_version
    except queue.Empty:
        proc.terminate()
        return 'Unknown'


def _opend_proc(q, ip=None, port=None):
    arg = {}
    if ip is not None:
        arg['host'] = ip
    if port is not None:
        arg['port'] = port

    quote_ctx = futu.OpenQuoteContext(**arg)
    ret, data = quote_ctx.get_global_state()
    if ret == futu.RET_OK:
        q.put(data['server_ver'])
    quote_ctx.close()


def _get_log_dir():
    if sys.platform.startswith('win'):
        return os.path.join(os.getenv("appdata"), 'com.futunn.FutuOpenD/Log')
    else:
        return os.path.join(os.environ['HOME'], '.com.futunn.FutuOpenD/Log')



# -*- coding: utf-8 -*-
import pandas as pd
from futu.common import RspHandlerBase
from futu.quote.quote_query import *


class StockQuoteHandlerBase(RspHandlerBase):
    """
    异步处理推送的订阅股票的报价。

    .. code:: python

        class StockQuoteTest(StockQuoteHandlerBase):
            def on_recv_rsp(self, rsp_str):
                ret_code, content = super(StockQuoteTest,self).on_recv_rsp(rsp_str)
                if ret_code != RET_OK:
                    print("StockQuoteTest: error, msg: %s" % content)
                    return RET_ERROR, content

                print("StockQuoteTest ", content) # StockQuoteTest自己的处理逻辑

                return RET_OK, content
    """
    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, quote_list = StockQuoteQuery.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, quote_list

    def on_recv_rsp(self, rsp_pb):
        """
        在收到实时报价推送后会回调到该函数，使用者需要在派生类中覆盖此方法

        注意该回调是在独立子线程中

        :param rsp_pb: 派生类中不需要直接处理该参数
        :return: 参见get_stock_quote的返回值
        """
        ret_code, content = self.parse_rsp_pb(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, content
        else:
            col_list = [
                'code', 'data_date', 'data_time', 'last_price', 'open_price',
                'high_price', 'low_price', 'prev_close_price', 'volume',
                'turnover', 'turnover_rate', 'amplitude', 'suspension',
                'listing_date', 'price_spread', 'dark_status', 'sec_status', 'strike_price',
                'contract_size', 'open_interest', 'implied_volatility',
                'premium', 'delta', 'gamma', 'vega', 'theta', 'rho',
                'net_open_interest', 'expiry_date_distance', 'contract_nominal_value', 
                'owner_lot_multiplier', 'option_area_type', 'contract_multiplier',
                'last_settle_price','position','position_change'
            ]

            col_list.extend(row[0] for row in pb_field_map_PreAfterMarketData_pre)
            col_list.extend(row[0] for row in pb_field_map_PreAfterMarketData_after)

            quote_frame_table = pd.DataFrame(content, columns=col_list)

            return RET_OK, quote_frame_table


class OrderBookHandlerBase(RspHandlerBase):
    """
    异步处理推送的实时摆盘。

    .. code:: python

        class OrderBookTest(OrderBookHandlerBase):
            def on_recv_rsp(self, rsp_str):
                ret_code, data = super(OrderBookTest,self).on_recv_rsp(rsp_str)
                if ret_code != RET_OK:
                    print("OrderBookTest: error, msg: %s" % data)
                    return RET_ERROR, data

                print("OrderBookTest ", data) # OrderBookTest自己的处理逻辑

                return RET_OK, content
    """
    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, order_book = OrderBookQuery.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, order_book

    def on_recv_rsp(self, rsp_pb):
        """
        在收到实摆盘数据推送后会回调到该函数，使用者需要在派生类中覆盖此方法

        注意该回调是在独立子线程中

        :param rsp_pb: 派生类中不需要直接处理该参数
        :return: 参见get_order_book的返回值
        """
        ret_code, content = self.parse_rsp_pb(rsp_pb)
        if ret_code == RET_OK:
            self.on_recv_log(content)

        return ret_code, content


class CurKlineHandlerBase(RspHandlerBase):
    """
    异步处理推送的k线数据。

    .. code:: python

        class CurKlineTest(CurKlineHandlerBase):
            def on_recv_rsp(self, rsp_str):
                ret_code, data = super(CurKlineTest,self).on_recv_rsp(rsp_str)
                if ret_code != RET_OK:
                    print("CurKlineTest: error, msg: %s" % data)
                    return RET_ERROR, data

                print("CurKlineTest ", data) # CurKlineTest自己的处理逻辑

                return RET_OK, content
    """
    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, kline_list = CurKlinePush.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, kline_list

    def on_recv_rsp(self, rsp_pb):
        """
        在收到实时k线数据推送后会回调到该函数，使用者需要在派生类中覆盖此方法

        注意该回调是在独立子线程中

        :param rsp_pb: 派生类中不需要直接处理该参数
        :return: 参见get_cur_kline的返回值
        """
        ret_code, content = self.parse_rsp_pb(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, content
        else:
            col_list = [
                'code', 'time_key', 'open', 'close', 'high', 'low', 'volume',
                'turnover', 'k_type', 'last_close'
            ]
            kline_frame_table = pd.DataFrame(content, columns=col_list)

            return RET_OK, kline_frame_table


class TickerHandlerBase(RspHandlerBase):
    """
    异步处理推送的逐笔数据。

    .. code:: python

        class TickerTest(TickerHandlerBase):
            def on_recv_rsp(self, rsp_str):
                ret_code, data = super(TickerTest,self).on_recv_rsp(rsp_str)
                if ret_code != RET_OK:
                    print("CurKlineTest: error, msg: %s" % data)
                    return RET_ERROR, data

                print("TickerTest ", data) # TickerTest自己的处理逻辑

                return RET_OK, content
    """
    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, ticker_list = TickerQuery.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, ticker_list

    def on_recv_rsp(self, rsp_pb):
        """
        在收到实时逐笔数据推送后会回调到该函数，使用者需要在派生类中覆盖此方法

        注意该回调是在独立子线程中

        :param rsp_pb: 派生类中不需要直接处理该参数
        :return: 参见get_rt_ticker的返回值
        """
        ret_code, content = self.parse_rsp_pb(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, content
        else:
            self.on_recv_log(content)
            col_list = [
                'code', 'time', 'price', 'volume', 'turnover',
                "ticker_direction", 'sequence', 'type', 'push_data_type',
            ]
            ticker_frame_table = pd.DataFrame(content, columns=col_list)

            return RET_OK, ticker_frame_table


class RTDataHandlerBase(RspHandlerBase):
    """
    异步处理推送的分时数据。

    .. code:: python

        class RTDataTest(RTDataHandlerBase):
            def on_recv_rsp(self, rsp_str):
                ret_code, data = super(RTDataTest,self).on_recv_rsp(rsp_str)
                if ret_code != RET_OK:
                    print("RTDataTest: error, msg: %s" % data)
                    return RET_ERROR, data

                print("RTDataTest ", data) # RTDataTest自己的处理逻辑

                return RET_OK, content
    """
    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, rt_data_list = RtDataQuery.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, rt_data_list

    def on_recv_rsp(self, rsp_pb):
        """
        在收到实时逐笔数据推送后会回调到该函数，使用者需要在派生类中覆盖此方法

        注意该回调是在独立子线程中

        :param rsp_pb: 派生类中不需要直接处理该参数
        :return: 参见get_rt_data的返回值
        """
        ret_code, content = self.parse_rsp_pb(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, content
        else:

            col_list = [
                'code', 'time', 'is_blank', 'opened_mins', 'cur_price',
                "last_close", 'avg_price', 'turnover', 'volume'
            ]
            rt_data_table = pd.DataFrame(content, columns=col_list)

            return RET_OK, rt_data_table


class BrokerHandlerBase(RspHandlerBase):
    """
    异步处理推送的经纪数据。

    .. code:: python

        class BrokerTest(BrokerHandlerBase):
            def on_recv_rsp(self, rsp_str):
                ret_code, data = super(BrokerTest,self).on_recv_rsp(rsp_str)
                if ret_code != RET_OK:
                    print("BrokerTest: error, msg: %s" % data)
                    return RET_ERROR, data

                print("BrokerTest ", data) # BrokerTest自己的处理逻辑

                return RET_OK, content
    """
    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, (stock_code, bid_content,
                        ask_content) = BrokerQueueQuery.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, (stock_code, bid_content, ask_content)

    def on_recv_rsp(self, rsp_pb):
        """
        在收到实时经纪数据推送后会回调到该函数，使用者需要在派生类中覆盖此方法

        注意该回调是在独立子线程中

        :param rsp_pb: 派生类中不需要直接处理该参数
        :return: 成功时返回(RET_OK, stock_code, [bid_frame_table, ask_frame_table]), 相关frame table含义见 get_broker_queue_ 的返回值说明

                 失败时返回(RET_ERROR, ERR_MSG, None)
        """
        ret_code, content = self.parse_rsp_pb(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, content, None
        else:
            self.on_recv_log(content)
            stock_code, bid_content, ask_content = content
            bid_list = [
                'code', 'bid_broker_id', 'bid_broker_name', 'bid_broker_pos', 'order_id', 'order_volume'
            ]
            ask_list = [
                'code', 'ask_broker_id', 'ask_broker_name', 'ask_broker_pos', 'order_id', 'order_volume'
            ]
            bid_frame_table = pd.DataFrame(bid_content, columns=bid_list)
            ask_frame_table = pd.DataFrame(ask_content, columns=ask_list)
            return ret_code, stock_code, [bid_frame_table, ask_frame_table]


class KeepAliveHandlerBase(RspHandlerBase):
    """Base class for handling KeepAlive"""

    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, alive_time = KeepAlive.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, alive_time

    def on_recv_rsp(self, rsp_pb):
        """receive response callback function"""

        ret_code, content = self.parse_rsp_pb(rsp_pb)

        return ret_code, content


class SysNotifyHandlerBase(RspHandlerBase):
    """sys notify"""

    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, content = SysNotifyPush.unpack_rsp(rsp_pb)

        return ret_code, content

    def on_recv_rsp(self, rsp_pb):
        """receive response callback function"""
        ret_code, content = self.parse_rsp_pb(rsp_pb)

        return ret_code, content

#
# class OrderDetailHandlerBase(RspHandlerBase):
#     def __init__(self):
#         super(OrderDetailHandlerBase, self).__init__()
#
#     def on_recv_rsp(self, rsp_pb):
#         """receive response callback function"""
#         ret_code, msg, data = OrderDetail.unpack_rsp(rsp_pb)
#
#         if ret_code != RET_OK:
#             return ret_code, msg
#         else:
#             return ret_code, data

class PriceReminderHandlerBase(RspHandlerBase):
    """
    异步处理推送的订阅股票的报价。

    .. code:: python

        class PriceReminderTest(PriceReminderHandlerBase):
            def on_recv_rsp(self, rsp_str):
                ret_code, content = super(PriceReminderTest,self).on_recv_rsp(rsp_str)
                if ret_code != RET_OK:
                    print("PriceReminderTest: error, msg: %s" % content)
                    return RET_ERROR, content

                print("PriceReminderTest ", content) # PriceReminderTest自己的处理逻辑

                return RET_OK, content
    """
    @classmethod
    def parse_rsp_pb(cls, rsp_pb):
        ret_code, msg, data = UpdatePriceReminder.unpack_rsp(rsp_pb)
        if ret_code != RET_OK:
            return ret_code, msg
        else:
            return RET_OK, data

    def on_recv_rsp(self, rsp_pb):
        """
        在收到实时报价推送后会回调到该函数，使用者需要在派生类中覆盖此方法

        注意该回调是在独立子线程中

        :param rsp_pb: 派生类中不需要直接处理该参数
        :return: 参见get_stock_quote的返回值
        """
        ret_code, content = self.parse_rsp_pb(rsp_pb)
        return ret_code, content


import errno
import datetime
import threading
from time import sleep
from futu.common.utils import *
from futu.quote.quote_query import parse_head
from .err import Err
from .sys_config import SysConfig
from .ft_logger import *

if IS_PY2:
    import selectors2 as selectors
    import Queue as queue
else:
    import queue
    import selectors


class ConnStatus:
    Start = 0
    Connecting = 1
    Connected = 2
    Closed = 3


class SyncReqRspInfo:
    def __init__(self):
        self.event = threading.Event()
        self.ret = RET_OK
        self.msg = ''
        self.data = None


class Connection:
    def __init__(self, conn_id, sock, addr, handler, is_encrypt):
        self._conn_id = conn_id
        self.opend_conn_id = 0
        self.sock = sock
        self.is_encrypt = is_encrypt
        self.handler = handler
        self._peer_addr = addr
        self.status = ConnStatus.Start
        self.keep_alive_interval = 10
        self.last_keep_alive_time = datetime.now()
        self.timeout = None
        self.start_time = None
        self.readbuf = bytearray()
        self.writebuf = bytearray()
        self.req_dict = {}  # ProtoInfo -> req time
        self.sync_req_dict = {}  # ProtoInfo -> SyncReqRspInfo

    @property
    def conn_id(self):
        return self._conn_id

    @property
    def peer_addr(self):
        return self._peer_addr

    def fileno(self):
        return self.sock.fileno


def is_socket_exception_wouldblock(e):
    has_errno = False
    if IS_PY2:
        if isinstance(e, IOError):
            has_errno = True
    else:
        if isinstance(e, OSError):
            has_errno = True

    if has_errno:
        if e.errno == errno.EWOULDBLOCK or e.errno == errno.EAGAIN or e.errno == errno.EINPROGRESS:
            return True
    return False


def make_ctrl_socks():
    LOCAL_HOST = '127.0.0.1'
    if IS_PY2:
        svr_sock = []
        lsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        def svr_sock_func():
            try:
                sock, _ = lsock.accept()
                svr_sock.append(sock)
            except Exception as e:
                logger.warning('Ctrl sock fail: {}'.format(str(e)))
        try:
            lsock.bind((LOCAL_HOST, 0))
            _, port = lsock.getsockname()[:2]
            lsock.listen(1)
            thread = threading.Thread(target=svr_sock_func)
            thread.start()
            client_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client_sock.settimeout(0.1)
            client_sock.connect((LOCAL_HOST, port))
            thread.join()
            return svr_sock[0], client_sock
        except Exception as e:
            logger.warning('Ctrl sock fail: {}'.format(str(e)))
            return None, None
        finally:
            lsock.close()
    else:
        return socket.socketpair()

class NetManager:
    _default_inst = None
    _default_inst_lock = threading.Lock()

    @classmethod
    def default(cls):
        with cls._default_inst_lock:
            if cls._default_inst is None:
                cls._default_inst = NetManager()
            return cls._default_inst

    def __init__(self):
        self._use_count = 0
        self._next_conn_id = 1
        self._lock = threading.RLock()
        self._mgr_lock = threading.Lock()  # Used to control start and stop
        self._create_all()

    def _close_all(self):
        for sel_key in list(self._selector.get_map().values()):
            self._selector.unregister(sel_key.fileobj)
            sel_key.fileobj.close()
        self._selector.close()
        self._selector = None
        if self._r_sock:
            self._r_sock.close()
            self._r_sock = None
        if self._w_sock:
            self._w_sock.close()
            self._w_sock = None

    def _create_all(self):
        self._selector = selectors.DefaultSelector()
        self._req_queue = queue.Queue()
        self._sync_req_timeout = 12
        self._thread = None
        now = datetime.now()
        self._last_activate_time = now
        self._last_check_req_time = now
        self._r_sock, self._w_sock = make_ctrl_socks()
        self._selector.register(self._r_sock, selectors.EVENT_READ)

    def connect(self, addr, handler, timeout, is_encrypt):
        with self._lock:
            conn_id = self._next_conn_id
            self._next_conn_id += 1

        def work():
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1024 * 1024)
            conn = Connection(conn_id, sock, addr, handler, is_encrypt)
            conn.status = ConnStatus.Connecting
            conn.start_time = datetime.now()
            conn.timeout = timeout
            sock.setblocking(False)
            self._selector.register(sock, selectors.EVENT_READ | selectors.EVENT_WRITE, conn)

            try:
                sock.connect(addr)
            except Exception as e:
                if not is_socket_exception_wouldblock(e):
                    conn.handler.on_error(conn.conn_id, str(e))
                    self.close(conn.conn_id)
                    return RET_ERROR, str(e), 0

            return RET_OK, '', conn_id

        self._req_queue.put(work)
        self._w_sock.send(b'1')
        return RET_OK, '', conn_id

    def poll(self):
        now = datetime.now()
        events = self._selector.select(0.02)
        for key, evt_mask in events:
            if key.fileobj == self._r_sock:
                self._r_sock.recv(1024)
                while True:
                    try:
                        work = self._req_queue.get(block=False)
                        work()
                    except queue.Empty:
                        break
                continue
            conn = key.data
            if evt_mask & selectors.EVENT_WRITE != 0:
                self._on_write(conn)

            if evt_mask & selectors.EVENT_READ != 0:
                self._on_read(conn)

        activate_elapsed_time = now - self._last_activate_time
        check_req_elapsed_time = now - self._last_check_req_time
        is_activate = activate_elapsed_time.total_seconds() >= 0.05
        is_check_req = check_req_elapsed_time.total_seconds() >= 0.1

        if is_activate or is_check_req:
            for key in list(self._selector.get_map().values()):
                if key.fileobj == self._r_sock:
                    continue
                conn = key.data
                if conn.status == ConnStatus.Connecting:
                    if is_activate:
                        self._check_connect_timeout(conn, now)
                elif conn.status == ConnStatus.Connected:
                    if is_activate:
                        conn.handler.on_activate(conn.conn_id, now)
                    if is_check_req:
                        self._check_req(conn, now)

        if is_activate:
            self._last_activate_time = now
        if is_check_req:
            self._last_check_req_time = now

    def _check_connect_timeout(self, conn, now):
        time_delta = now - conn.start_time
        if conn.timeout is not None and conn.timeout > 0 and time_delta.total_seconds() >= conn.timeout:
            self._on_connect_timeout(conn)

    def _check_req(self, conn, now):
        """

        :param conn:
        :type conn: Connection
        :param now:
        :type now: datetime
        :return:
        """
        req_dict = dict(conn.req_dict.items())
        for proto_info, req_time in req_dict.items():  # type: ProtoInfo, datetime
            elapsed_time = now - req_time
            if elapsed_time.total_seconds() >= self._sync_req_timeout:
                self._on_packet(conn, proto_info._asdict(), Err.Timeout.code, Err.Timeout.text, None)

    def _thread_func(self):
        while True:
            with self._lock:
                if not self.is_alive():
                    self._thread = None
                    break
            self.poll()

    def start(self):
        """
        Should be called from main thread
        :return:
        """
        with self._mgr_lock:
            with self._lock:
                self._use_count += 1

            if self._thread is None:
                self._create_all()
                self._thread = threading.Thread(target=self._thread_func)
                self._thread.setDaemon(SysConfig.get_all_thread_daemon())
                self._thread.start()

    def stop(self):
        with self._mgr_lock:
            with self._lock:
                self._use_count = max(self._use_count - 1, 0)

    def is_alive(self):
        with self._lock:
            return self._use_count > 0

    def do_send(self, conn_id, proto_info, data):
        logger.debug2(FTLog.ONLY_FILE, 'Send: conn_id={}; proto_id={}; serial_no={}; total_len={};'.format(conn_id, proto_info.proto_id,
                                                                                         proto_info.serial_no,
                                                                                         len(data)))
        now = datetime.now()
        ret_code = RET_OK
        msg = ''
        conn = self._get_conn(conn_id)  # type: Connection
        sync_req_rsp = None
        if not conn:
            logger.debug(
                FTLog.make_log_msg('Send fail', conn_id=conn_id, proto_id=proto_info.proto_id, serial_no=proto_info.serial_no,
                             msg=Err.ConnectionLost.text))
            ret_code, msg = RET_ERROR, Err.ConnectionLost.text
        else:
            sync_req_rsp = conn.sync_req_dict.get(proto_info, None)

        if ret_code != RET_OK:
            return ret_code, msg

        if conn.status != ConnStatus.Connected:
            ret_code, msg = RET_ERROR, Err.NotConnected.text

        if ret_code != RET_OK:
            logger.warning(FTLog.make_log_msg('Send fail', proto_id=proto_info.proto_id, serial_no=proto_info.serial_no,
                                        conn_id=conn_id, msg=msg))
            if sync_req_rsp:
                sync_req_rsp.ret, sync_req_rsp.msg = RET_ERROR, msg
                sync_req_rsp.event.set()

            return ret_code, msg

        conn.req_dict[proto_info] = now
        size = 0
        try:
            if len(conn.writebuf) > 0:
                conn.writebuf.extend(data)
            else:
                size = conn.sock.send(data)
        except Exception as e:
            if is_socket_exception_wouldblock(e):
                pass
            else:
                ret_code, msg = RET_ERROR, str(e)

        if size > 0 and size < len(data):
            conn.writebuf.extend(data[size:])
            self._watch_write(conn, True)

        if ret_code != RET_OK:
            logger.warning(FTLog.make_log_msg('Send error', conn_id=conn_id, msg=msg))
            if sync_req_rsp:
                sync_req_rsp.ret, sync_req_rsp.msg = RET_ERROR, msg
                sync_req_rsp.event.set()
            return ret_code, msg

        return RET_OK, ''

    def send(self, conn_id, data):
        """

        :param conn_id:
        :param data:
        :return:
        """
        proto_info = self._parse_req_head_proto_info(data)

        def work():
            self.do_send(conn_id, proto_info, data)

        self._req_queue.put(work)
        self._w_sock.send(b'1')
        return RET_OK, None

    def close(self, conn_id):
        def work():
            conn = self._get_conn(conn_id)  # type: Connection
            if not conn:
                return
            if conn.sock is None:
                return
            self._watch_read(conn, False)
            self._watch_write(conn, False)
            conn.sock.close()
            conn.sock = None
            conn.status = ConnStatus.Closed
            for proto_info, sync_req_rsp in conn.sync_req_dict.items():  # type: ProtoInfo, SyncReqRspInfo
                sync_req_rsp.ret = RET_ERROR
                sync_req_rsp.msg = Err.ConnectionClosed.text
                sync_req_rsp.event.set()
            logger.info("Close: conn_id={}".format(conn_id))

        self._req_queue.put(work)
        self._w_sock.send(b'1')

    def _watch_read(self, conn, is_watch):
        try:
            sel_key = self._selector.get_key(conn.sock)
        except KeyError:
            return

        if is_watch:
            new_event = sel_key.events | selectors.EVENT_READ
        else:
            new_event = sel_key.events & (~selectors.EVENT_READ)

        if new_event != 0:
            self._selector.modify(conn.sock, new_event, conn)
        else:
            self._selector.unregister(conn.sock)

    def _watch_write(self, conn, is_watch):
        try:
            sel_key = self._selector.get_key(conn.sock)
        except KeyError:
            return

        if is_watch:
            new_event = sel_key.events | selectors.EVENT_WRITE
        else:
            new_event = sel_key.events & (~selectors.EVENT_WRITE)

        if new_event != 0:
            self._selector.modify(conn.sock, new_event, conn)
        else:
            self._selector.unregister(conn.sock)

    def sync_query(self, conn_id, req_str):
        head_dict = self._parse_req_head(req_str)
        proto_info = ProtoInfo(head_dict['proto_id'], head_dict['serial_no'])
        rsp_info = SyncReqRspInfo()

        def work():
            conn = self._get_conn(conn_id)  # type: Connection
            ret, msg = RET_OK, ''
            if not conn:
                ret = RET_ERROR
                msg = Err.ConnectionLost.text
            else:
                conn.sync_req_dict[proto_info] = rsp_info
                self.do_send(conn_id, proto_info, req_str)

            if ret != RET_OK:
                rsp_info.ret = ret
                rsp_info.msg = msg
                rsp_info.event.set()

        self._req_queue.put(work)
        self._w_sock.send(b'1')

        rsp_info.event.wait()
        return rsp_info.ret, rsp_info.msg, rsp_info.data

    def _parse_req_head(self, req_str):
        head_len = get_message_head_len()
        req_head_dict = parse_head(req_str[:head_len])
        return req_head_dict

    def _parse_req_head_proto_info(selfs, req_str):
        head_len = get_message_head_len()
        proto_info = parse_proto_info(req_str[:head_len])
        return proto_info

    def _get_conn(self, conn_id):
        with self._lock:
            for sock, sel_key in self._selector.get_map().items():
                if sel_key.fileobj == self._r_sock:
                    continue
                conn = sel_key.data
                if conn.conn_id == conn_id:
                    return conn
            return None

    def _on_read(self, conn):
        start_time = time.time()
        recv_len = 0
        buf_len = 0
        packet_count = 0

        if conn.status == ConnStatus.Closed:
            return

        err = None
        is_closed = False
        try:
            data = conn.sock.recv(128 * 1024)
            if data == b'':
                is_closed = True
            else:
                conn.readbuf.extend(data)
                recv_len = len(data)
                buf_len = len(conn.readbuf)

        except Exception as e:
            if not is_socket_exception_wouldblock(e):
                err = str(e)

        while len(conn.readbuf) > 0:
            head_len = get_message_head_len()
            if len(conn.readbuf) < head_len:
                break
            head_dict = parse_head(conn.readbuf[:head_len])
            body_len = head_dict['body_len']
            if len(conn.readbuf) < head_len + body_len:
                break

            rsp_body = conn.readbuf[head_len:head_len+body_len]
            del conn.readbuf[:head_len+body_len]
            packet_count += 1
            self._on_packet(conn, head_dict, Err.Ok.code, '', rsp_body)

        if is_closed:
            self.close(conn.conn_id)
            conn.handler.on_error(conn.conn_id, Err.ConnectionClosed.text)
        elif err:
            self.close(conn.conn_id)
            conn.handler.on_error(conn.conn_id, err)
        # end_time = time.time()
        # logger.debug2(FTLog.ONLY_FILE, 'conn_id={}; elapsed={}; recv_len={}; buf_len={}; packet={};'.format(conn.conn_id, end_time-start_time, recv_len, buf_len, packet_count))

    def _on_write(self, conn):
        if conn.status == ConnStatus.Closed:
            return
        elif conn.status == ConnStatus.Connecting:
            err = conn.sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
            self._watch_write(conn, False)
            if err != 0:
                conn.handler.on_error(conn.conn_id, errno.errorcode[err])
            else:
                conn.status = ConnStatus.Connected
                conn.handler.on_connected(conn.conn_id)
            return

        err = None

        size = 0
        try:
            if len(conn.writebuf) > 0:
                size = conn.sock.send(conn.writebuf)
        except Exception as e:
            if not is_socket_exception_wouldblock(e):
                err = str(e)

        if size > 0:
            del conn.writebuf[:size]

        if len(conn.writebuf) == 0:
            self._watch_write(conn, False)

        if err:
            self.close(conn.conn_id)
            conn.handler.on_error(conn.conn_id, err)

    def _on_connect_timeout(self, conn):
        conn.handler.on_connect_timeout(conn.conn_id)

    def _on_packet(self, conn, head_dict, err_code, msg, rsp_body_data):
        """

        :param conn:
        :type conn: Connection
        :param head_dict:
        :param err_code:
        :param msg:
        :param rsp_body_data:
        :return:
        """
        proto_info = ProtoInfo(head_dict['proto_id'], head_dict['serial_no'])
        rsp_pb = None
        if err_code == Err.Ok.code:
            ret_decrypt, msg_decrypt, rsp_body = decrypt_rsp_body(rsp_body_data, head_dict, conn.opend_conn_id, conn.is_encrypt)
            if ret_decrypt == RET_OK:
                rsp_pb = binary2pb(rsp_body, head_dict['proto_id'], head_dict['proto_fmt_type'])
            else:
                err_code = Err.PacketDataErr.code
                msg = msg_decrypt
                rsp_pb = None

        log_msg = 'Recv: conn_id={}; proto_id={}; serial_no={}; data_len={}; msg={};'.format(conn.conn_id,
                                                                                             proto_info.proto_id,
                                                                                             proto_info.serial_no,
                                                                                             len(
                                                                                                 rsp_body_data) if rsp_body_data else 0,
                                                                                             msg)
        if err_code == Err.Ok.code:
            logger.debug2(FTLog.ONLY_FILE, log_msg)
        else:
            logger.warning(log_msg)

        ret_code = RET_OK if err_code == Err.Ok.code else RET_ERROR
        sync_rsp_info = conn.sync_req_dict.get(proto_info, None)  # type: SyncReqRspInfo
        conn.req_dict.pop(proto_info, None)
        if sync_rsp_info:
            sync_rsp_info.ret, sync_rsp_info.msg, sync_rsp_info.data = ret_code, msg, rsp_pb
            sync_rsp_info.event.set()
            conn.sync_req_dict.pop(proto_info)
        else:
            conn.handler.on_packet(conn.conn_id, proto_info, ret_code, msg, rsp_pb)

    @staticmethod
    def extract_rsp_pb(opend_conn_id, head_dict, rsp_body):
        ret, msg, rsp = decrypt_rsp_body(rsp_body, head_dict, opend_conn_id)
        if ret == RET_OK:
            rsp_pb = binary2pb(rsp_body, head_dict['proto_id'], head_dict['proto_fmt_type'])
        else:
            rsp_pb = None
        return ret, msg, rsp_pb

    def set_conn_info(self, conn_id, info):
        with self._lock:
            conn = self._get_conn(conn_id)
            if conn is not None:
                conn.opend_conn_id = info.get('conn_id', conn.opend_conn_id)
                conn.keep_alive_interval = info.get('keep_alive_interval', conn.keep_alive_interval)
            else:
                return RET_ERROR, Err.ConnectionLost.text
        return RET_OK, ''



# -*- coding: utf-8 -*-
import talib
import math
import datetime
import logging
import futu as ft


class MACD(object):
    """
    A simple MACD strategy
    """
    # API parameter setting
    api_svr_ip = '127.0.0.1'  # 账户登录的牛牛客户端PC的IP, 本机默认为127.0.0.1
    api_svr_port = 11111  # 富途牛牛端口，默认为11111
    unlock_password = "123456"  # 美股和港股交易解锁密码
    trade_env = ft.TrdEnv.SIMULATE

    def __init__(self, stock, short_period, long_period, smooth_period,
                 observation):
        """
        Constructor
        """
        self.stock = stock
        self.short_period = short_period
        self.long_period = long_period
        self.smooth_period = smooth_period
        self.observation = observation
        self.quote_ctx, self.trade_ctx = self.context_setting()

    def close(self):
        self.quote_ctx.close()
        self.trade_ctx.close()

    def context_setting(self):
        """
        API trading and quote context setting
        :returns: trade context, quote context
        """
        if self.unlock_password == "":
            raise Exception("请先配置交易解锁密码! password: {}".format(
                self.unlock_password))

        quote_ctx = ft.OpenQuoteContext(
            host=self.api_svr_ip, port=self.api_svr_port)

        if 'HK.' in self.stock:
            trade_ctx = ft.OpenHKTradeContext(host=self.api_svr_ip, port=self.api_svr_port)
        elif 'US.' in self.stock:
            trade_ctx = ft.OpenUSTradeContext(host=self.api_svr_ip, port=self.api_svr_port)
        else:
            raise Exception("不支持的stock: {}".format(self.stock))

        if self.trade_env == ft.TrdEnv.REAL:
            ret_code, ret_data = trade_ctx.unlock_trade(
                self.unlock_password)
            if ret_code == ft.RET_OK:
                print('解锁交易成功!')
            else:
                raise Exception("请求交易解锁失败: {}".format(ret_data))
        else:
            print('解锁交易成功!')

        return quote_ctx, trade_ctx

    def handle_data(self):
        """
        handle stock data for trading signal, and make order
        """
        # 读取历史数据，使用sma方式计算均线准确度和数据长度无关，但是在使用ema方式计算均线时建议将历史数据窗口适当放大，结果会更加准确
        today = datetime.datetime.today()
        pre_day = (today - datetime.timedelta(days=self.observation)
                   ).strftime('%Y-%m-%d')
        end_dt = today.strftime('%Y-%m-%d')
        ret_code, prices, page_req_key = self.quote_ctx.request_history_kline(self.stock, start=pre_day, end=end_dt)
        if ret_code != ft.RET_OK:
            print("request_history_kline fail: {}".format(prices))
            return

        # 用talib计算MACD取值，得到三个时间序列数组，分别为 macd, signal 和 hist
        # macd 是长短均线的差值，signal 是 macd 的均线
        # 使用 macd 策略有几种不同的方法，我们这里采用 macd 线突破 signal 线的判断方法
        macd, signal, hist = talib.MACD(prices['close'].values,
                                        self.short_period, self.long_period,
                                        self.smooth_period)

        # 如果macd从上往下跌破macd_signal
        if macd[-1] < signal[-1] and macd[-2] > signal[-2]:
            # 计算现在portfolio中股票的仓位
            ret_code, data = self.trade_ctx.position_list_query(
                trd_env=self.trade_env)

            if ret_code != ft.RET_OK:
                raise Exception('账户信息获取失败: {}'.format(data))
            pos_info = data.set_index('code')

            cur_pos = int(pos_info['qty'][self.stock])
            # 进行清仓
            if cur_pos > 0:
                ret_code, data = self.quote_ctx.get_market_snapshot(
                    [self.stock])
                if ret_code != 0:
                    raise Exception('市场快照数据获取异常 {}'.format(data))
                cur_price = data['last_price'][0]
                ret_code, ret_data = self.trade_ctx.place_order(
                    price=cur_price,
                    qty=cur_pos,
                    code=self.stock,
                    trd_side=ft.TrdSide.SELL,
                    order_type=ft.OrderType.NORMAL,
                    trd_env=self.trade_env)
                if ret_code == ft.RET_OK:
                    print('stop_loss MAKE SELL ORDER\n\tcode = {} price = {} quantity = {}'
                          .format(self.stock, cur_price, cur_pos))
                else:
                    print('stop_loss: MAKE SELL ORDER FAILURE: {}'.format(ret_data))

        # 如果短均线从下往上突破长均线，为入场信号
        if macd[-1] > signal[-1] and macd[-2] < signal[-2]:
            # 满仓入股
            ret_code, acc_info = self.trade_ctx.accinfo_query(
                trd_env=self.trade_env)
            if ret_code != 0:
                raise Exception('账户信息获取失败! 请重试: {}'.format(acc_info))

            ret_code, snapshot = self.quote_ctx.get_market_snapshot(
                [self.stock])
            if ret_code != 0:
                raise Exception('市场快照数据获取异常 {}'.format(snapshot))
            lot_size = snapshot['lot_size'][0]
            cur_price = snapshot['last_price'][0]
            cash = acc_info['power'][0]  # 购买力
            qty = int(math.floor(cash / cur_price))
            qty = qty // lot_size * lot_size

            ret_code, ret_data = self.trade_ctx.place_order(
                price=cur_price,
                qty=qty,
                code=self.stock,
                trd_side=ft.TrdSide.BUY,
                order_type=ft.OrderType.NORMAL,
                trd_env=self.trade_env)
            if not ret_code:
                print(
                    'stop_loss MAKE BUY ORDER\n\tcode = {} price = {} quantity = {}'
                    .format(self.stock, cur_price, qty))
            else:
                print('stop_loss: MAKE BUY ORDER FAILURE: {}'.format(ret_data))


if __name__ == "__main__":
    SHORT_PERIOD = 12
    LONG_PERIOD = 26
    SMOOTH_PERIOD = 9
    OBSERVATION = 100

    STOCK = "HK.00123"

    test = MACD(STOCK, SHORT_PERIOD, LONG_PERIOD, SMOOTH_PERIOD, OBSERVATION)
    test.handle_data()
    test.close()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Qot_UpdateTicker.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import Common_pb2 as Common__pb2
import Qot_Common_pb2 as Qot__Common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='Qot_UpdateTicker.proto',
  package='Qot_UpdateTicker',
  syntax='proto2',
  serialized_pb=_b('\n\x16Qot_UpdateTicker.proto\x12\x10Qot_UpdateTicker\x1a\x0c\x43ommon.proto\x1a\x10Qot_Common.proto\"U\n\x03S2C\x12&\n\x08security\x18\x01 \x02(\x0b\x32\x14.Qot_Common.Security\x12&\n\ntickerList\x18\x02 \x03(\x0b\x32\x12.Qot_Common.Ticker\"f\n\x08Response\x12\x15\n\x07retType\x18\x01 \x02(\x05:\x04-400\x12\x0e\n\x06retMsg\x18\x02 \x01(\t\x12\x0f\n\x07\x65rrCode\x18\x03 \x01(\x05\x12\"\n\x03s2c\x18\x04 \x01(\x0b\x32\x15.Qot_UpdateTicker.S2CBF\n\x13\x63om.futu.openapi.pbZ/github.com/futuopen/ftapi4go/pb/qotupdateticker')
  ,
  dependencies=[Common__pb2.DESCRIPTOR,Qot__Common__pb2.DESCRIPTOR,])




_S2C = _descriptor.Descriptor(
  name='S2C',
  full_name='Qot_UpdateTicker.S2C',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='security', full_name='Qot_UpdateTicker.S2C.security', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tickerList', full_name='Qot_UpdateTicker.S2C.tickerList', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=76,
  serialized_end=161,
)


_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='Qot_UpdateTicker.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='retType', full_name='Qot_UpdateTicker.Response.retType', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=-400,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='retMsg', full_name='Qot_UpdateTicker.Response.retMsg', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='errCode', full_name='Qot_UpdateTicker.Response.errCode', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='s2c', full_name='Qot_UpdateTicker.Response.s2c', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=163,
  serialized_end=265,
)

_S2C.fields_by_name['security'].message_type = Qot__Common__pb2._SECURITY
_S2C.fields_by_name['tickerList'].message_type = Qot__Common__pb2._TICKER
_RESPONSE.fields_by_name['s2c'].message_type = _S2C
DESCRIPTOR.message_types_by_name['S2C'] = _S2C
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

S2C = _reflection.GeneratedProtocolMessageType('S2C', (_message.Message,), dict(
  DESCRIPTOR = _S2C,
  __module__ = 'Qot_UpdateTicker_pb2'
  # @@protoc_insertion_point(class_scope:Qot_UpdateTicker.S2C)
  ))
_sym_db.RegisterMessage(S2C)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), dict(
  DESCRIPTOR = _RESPONSE,
  __module__ = 'Qot_UpdateTicker_pb2'
  # @@protoc_insertion_point(class_scope:Qot_UpdateTicker.Response)
  ))
_sym_db.RegisterMessage(Response)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\023com.futu.openapi.pbZ/github.com/futuopen/ftapi4go/pb/qotupdateticker'))
# @@protoc_insertion_point(module_scope)



import os


class FutuTemplate(object):
    def __init__(self, template_filename):
        self.local_path = os.path.dirname(os.path.realpath(__file__))
        self.template = dict()
        with open(os.path.join(self.local_path, template_filename), 'r', encoding='UTF-8') as load_f:
            begin_name = template_code = ""

            while True:
                line = load_f.readline()
                if not line:
                    break
                if len(line) > 10:
                    prefix = line[:10]
                else:
                    prefix = ""
                if prefix == "@@@@@@@@@@":  # 专用注释
                    continue
                if prefix == "##########":
                    if begin_name == "":
                        end = line.find('#', 11)
                        begin_name = line[10:end]
                        template_code = ""
                    else:
                        self.template[begin_name] = template_code
                        begin_name = template_code = ""
                    continue
                if begin_name != "":
                    template_code += line\


    @classmethod
    def code_add_space(cls, code, space_count=0, space_str="    "):  # 逐行加空格
        """传入字符串，逐行加入空格，符合python规范"""
        if space_count == 0:
            return code

        ls = code.split('\n')
        ret_code = space_line = ""
        for i in range(space_count):
            space_line += space_str

        for s in ls:
            if len(s) > 0:
                ret_code += (space_line + s + "\n")
            else:
                ret_code += "\n"

        return ret_code[:-1]

    def __getitem__(self, key):
        return self.template[key]

    def get(self, key, space_count=0):
        str_code = self[key]
        return self.code_add_space(str_code, space_count)














# -*- coding: utf-8 -*-

from collections import namedtuple

_ErrField = namedtuple('_ErrField', ('code', 'text'))


class Err:
    """
    0 ~ 999 通用错误
    1000 ~ 1999 行情错误
    2000 ~ 2999 交易错误
    """
    Ok = _ErrField(0, 'Ok')
    ConnectionLost = _ErrField(1, 'Connection lost')
    Timeout = _ErrField(2, 'Timeout')
    NotConnected = _ErrField(3, 'Not connected')
    PacketDataErr = _ErrField(4, 'Packet data error')
    ConnectionClosed = _ErrField(5, 'Connection closed')
    ParamErr = _ErrField(6, 'Parameter error')
    NotSetRSAFile = _ErrField(7, 'Conn is encrypted, but no RSA private key file is set. Call SysConfig.set_init_rsa_file.')
    RsaErr = _ErrField(8, 'RSA key is invalid')
    WrongType = _ErrField(9, 'Wrong type')
    WrongValue = _ErrField(10, 'Wrong value')
    NoNeedUnlock = _ErrField(2000, 'No need to unlock, because REAL trade is not supported in this market')
    NoAccForSecurityFirm = _ErrField(2001, 'Something wrong with the specified securities firm ')


def _make_kwargs_str(**kwargs):
    msg = ''
    if len(kwargs) > 0:
        msg = ':'
        for k, v in kwargs.items():
            msg += ' {0}={1};'.format(k, v)
    return msg


def make_msg(err, msg_=None, **kwargs):
    msg_ = '' if msg_ is None else ': ' + msg_
    if isinstance(err, str):
        return err + msg_ + _make_kwargs_str(**kwargs)
    else:
        return err.text + msg_ + _make_kwargs_str(**kwargs)

def make_wrong_type_msg(var_name, right_type):
    return '{}. {} should be {}'.format(Err.WrongType.text, var_name, right_type)

def make_wrong_value_msg(var_name, right_value):
    return '{}. {} should be {}'.format(Err.WrongValue.text, var_name, right_value)


def make_wrong_value_msg_utf8_len_le(var_name, max_len):
    return 'Error variable. The {} parameter cannot be longer than {} bytes after being converted to utf8'.format(
        var_name,
        max_len)



# -*- coding: utf-8 -*-

import hashlib, json, os, sys, socket, traceback, time, struct, collections
from datetime import datetime, timedelta
from struct import calcsize
from google.protobuf.json_format import MessageToJson
from threading import RLock
from futu.common.conn_mng import *
from futu.common.sys_config import *
from futu.common.pbjson import json2pb


ProtoInfo = collections.namedtuple('ProtoInfo', ['proto_id', 'serial_no'])


def get_message_head_len():
    return calcsize(MESSAGE_HEAD_FMT)


def check_date_str_format(s, default_time="00:00:00"):
    """Check the format of date string"""
    try:
        str_fmt = s
        if ":" not in s:
            str_fmt = '{} {}'.format(s, default_time)

        dt_obj = datetime.strptime(str_fmt, "%Y-%m-%d %H:%M:%S")

        return RET_OK, dt_obj

    except ValueError:
        error_str = ERROR_STR_PREFIX + "wrong time or time format"
        return RET_ERROR, error_str


def normalize_date_format(date_str, default_time="00:00:00"):
    """normalize the format of data"""
    ret_code, ret_data = check_date_str_format(date_str, default_time)
    if ret_code != RET_OK:
        return ret_code, ret_data

    return RET_OK, ret_data.strftime("%Y-%m-%d %H:%M:%S")


def normalize_start_end_date(start, end, delta_days=0, default_time_start="00:00:00", default_time_end="23:59:59", prefer_end_now=True):
    """

    :param start:
    :param end:
    :param delta_days:
    :param default_time_start:
    :param default_time_end:
    :param prefer_end_now: 为True时，当start和end都为None时，end设为当前时间，为False则start设为当前时间
    :return:
    """
    if start is not None and is_str(start) is False:
        error_str = ERROR_STR_PREFIX + "the type of start param is wrong"
        return RET_ERROR, error_str, None, None

    if end is not None and is_str(end) is False:
        error_str = ERROR_STR_PREFIX + "the type of end param is wrong"
        return RET_ERROR, error_str, None, None

    dt_start = None
    dt_end = None
    delta = timedelta(days=delta_days)
    hour_end, min_end, sec_end = [int(x) for x in default_time_end.split(':')]
    hour_start, min_start, sec_start = [int(x) for x in default_time_start.split(':')]

    if start:
        ret_code, ret_data = check_date_str_format(start, default_time_start)
        if ret_code != RET_OK:
            return ret_code, ret_data, start, end
        dt_start = ret_data

    if end:
        ret_code, ret_data = check_date_str_format(end, default_time_end)
        if ret_code != RET_OK:
            return ret_code, ret_data, start, end
        dt_end = ret_data

    if end and not start:
        dt_tmp = dt_end - delta
        dt_start = datetime(year=dt_tmp.year, month=dt_tmp.month, day=dt_tmp.day, hour=hour_start, minute=min_start, second=sec_start)

    if start and not end:
        dt_tmp = dt_start + delta
        dt_end = datetime(year=dt_tmp.year, month=dt_tmp.month, day=dt_tmp.day, hour=hour_end, minute=min_end, second=sec_end)

    if not start and not end:
        if prefer_end_now:
            dt_now = datetime.now()
            dt_end = datetime(year=dt_now.year, month=dt_now.month, day=dt_now.day, hour=hour_end, minute=min_end, second=sec_end)
            dt_tmp = dt_end - delta
            dt_start = datetime(year=dt_tmp.year, month=dt_tmp.month, day=dt_tmp.day, hour=hour_start, minute=min_start, second=sec_start)
        else:
            dt_now = datetime.now()
            dt_start = datetime(year=dt_now.year, month=dt_now.month, day=dt_now.day, hour=hour_start, minute=min_start,
                              second=sec_start)
            dt_tmp = dt_start + delta
            dt_end = datetime(year=dt_tmp.year, month=dt_tmp.month, day=dt_tmp.day, hour=hour_end, minute=min_end,
                                second=sec_end)

    start = dt_start.strftime("%Y-%m-%d %H:%M:%S")
    end = dt_end.strftime("%Y-%m-%d %H:%M:%S")

    return RET_OK, '', start, end


def is_str(s):
    if IS_PY2:
        return isinstance(s, str) or isinstance(s, unicode)
    else:
        return isinstance(s, str)


def extract_pls_rsp(rsp_str):
    """Extract the response of PLS"""
    try:
        rsp = json.loads(rsp_str)
    except ValueError:
        traceback.print_exc()
        err = sys.exc_info()[1]
        err_str = ERROR_STR_PREFIX + str(err)
        return RET_ERROR, err_str, None

    error_code = int(rsp['retType'])

    if error_code != 1:
        error_str = ERROR_STR_PREFIX + rsp['retMsg']
        return RET_ERROR, error_str, None

    return RET_OK, "", rsp


def split_stock_str(stock_str_param):
    """split the stock string"""
    stock_str = str(stock_str_param)

    split_loc = stock_str.find(".")
    '''do not use the built-in split function in python.
    The built-in function cannot handle some stock strings correctly.
    for instance, US..DJI, where the dot . itself is a part of original code'''
    if 0 <= split_loc < len(stock_str) - 1 and Market.if_has_key(stock_str[0:split_loc]):
        market_str = stock_str[0:split_loc]
        _, market_code = Market.to_number(market_str)
        partial_stock_str = stock_str[split_loc + 1:]
        return RET_OK, (market_code, partial_stock_str)

    else:
        error_str = ERROR_STR_PREFIX + "format of code %s is wrong. (US.AAPL, HK.00700, SZ.000001)" % stock_str
        return RET_ERROR, error_str


def merge_qot_mkt_stock_str(qot_mkt, partial_stock_str):
    """
    Merge the string of stocks
    :param market: market code
    :param partial_stock_str: original stock code string. i.e. "AAPL","00700", "000001"
    :return: unified representation of a stock code. i.e. "US.AAPL", "HK.00700", "SZ.000001"

    """
    market_str = Market.to_string2(qot_mkt)
    stock_str = '.'.join([market_str, partial_stock_str])
    return stock_str


def merge_trd_mkt_stock_str(trd_sec_mkt, partial_stock_str):
    """
    Merge the string of stocks
    :param market: market code
    :param partial_stock_str: original stock code string. i.e. "AAPL","00700", "000001"
    :return: unified representation of a stock code. i.e. "US.AAPL", "HK.00700", "SZ.000001"

    """
    mkt_qot = Market.NONE
    if trd_sec_mkt == Trd_Common_pb2.TrdSecMarket_HK:
        mkt_qot = Market.HK
    elif trd_sec_mkt == Trd_Common_pb2.TrdSecMarket_CN_SH:
        mkt_qot = Market.SH
    elif trd_sec_mkt == Trd_Common_pb2.TrdSecMarket_CN_SZ:
        mkt_qot = Market.SZ
    elif trd_sec_mkt == Trd_Common_pb2.TrdSecMarket_US:
        mkt_qot = Market.US
    _, mkt = Market.to_number(mkt_qot)
    return merge_qot_mkt_stock_str(mkt, partial_stock_str)


def str2binary(s):
    """
    Transfer string to binary
    :param s: string content to be transformed to binary
    :return: binary
    """
    return s.encode('utf-8')


def is_str(obj):
    if sys.version_info.major == 3:
        return isinstance(obj, str) or isinstance(obj, bytes)
    else:
        return isinstance(obj, basestring)


def price_to_str_int1000(price):
    return str(int(round(float(price) * 1000,
                         0))) if str(price) is not '' else ''


# 1000*int price to float val
def int1000_price_to_float(price):
    return round(float(price) / 1000.0,
                 3) if str(price) is not '' else float(0)


# 10^9 int price to float val
def int10_9_price_to_float(price):
    return round(float(price) / float(10**9),
                 3) if str(price) is not '' else float(0)


# list 参数除重及规整
def unique_and_normalize_list(lst):
    ret = []
    if not lst:
        return ret
    tmp = lst if isinstance(lst, list) else [lst]
    [ret.append(x) for x in tmp if x not in ret]
    return ret


def md5_transform(raw_str):
    h1 = hashlib.md5()
    h1.update(raw_str.encode(encoding='utf-8'))
    return h1.hexdigest()


g_unique_id = int(time.time() % 10000)
g_unique_lock = RLock()
def get_unique_id32():
    global g_unique_id
    with g_unique_lock:
        g_unique_id += 1
        if g_unique_id >= 4294967295:
            g_unique_id = int(time.time() % 10000)
        ret_id = g_unique_id
    return ret_id


class ProtobufMap(dict):
    created_protobuf_map = {}

    def __init__(self):

        """ InitConnect = 1001  # 初始化连接 """
        from futu.common.pb.InitConnect_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.InitConnect] = Response()

        """ GetGlobalState = 1002  # 获取全局状态 """
        from futu.common.pb.GetGlobalState_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.GetGlobalState] = Response()

        """ Notify = 1003  # 通知推送 """
        from futu.common.pb.Notify_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Notify] = Response()

        """ KeepAlive = 1004  # 通知推送 """
        from futu.common.pb.KeepAlive_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.KeepAlive] = Response()

        """ GetUserInfo = 1005  # 获取全局状态 """
        from futu.common.pb.GetUserInfo_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.GetUserInfo] = Response()

        """ GetUserInfo = 1006  # 获取用户信息 """
        from futu.common.pb.Verification_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Verification] = Response()

        """ GetUserInfo = 1007  # 获取延迟统计 """
        from futu.common.pb.GetDelayStatistics_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.GetDelayStatistics] = Response()

        """ TestCmd = 1008  # 测试命令 """
        from futu.common.pb.TestCmd_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.TestCmd] = Response()

        """ Trd_GetAccList = 2001  # 获取业务账户列表 """
        from futu.common.pb.Trd_GetAccList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetAccList] = Response()

        """ Trd_UnlockTrade = 2005  # 解锁或锁定交易 """
        from futu.common.pb.Trd_UnlockTrade_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_UnlockTrade] = Response()

        """ Trd_SubAccPush = 2008  # 订阅业务账户的交易推送数据 """
        from futu.common.pb.Trd_SubAccPush_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_SubAccPush] = Response()

        """  Trd_GetFunds = 2101  # 获取账户资金 """
        from futu.common.pb.Trd_GetFunds_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetFunds] = Response()

        """ Trd_GetPositionList = 2102  # 获取账户持仓 """
        from futu.common.pb.Trd_GetPositionList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetPositionList] = Response()

        """ Trd_GetOrderList = 2201  # 获取订单列表 """
        from futu.common.pb.Trd_GetOrderList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetOrderList] = Response()

        """ Trd_PlaceOrder = 2202  # 下单 """
        from futu.common.pb.Trd_PlaceOrder_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_PlaceOrder] = Response()

        """ Trd_ModifyOrder = 2205  # 修改订单 """
        from futu.common.pb.Trd_ModifyOrder_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_ModifyOrder] = Response()

        """ Trd_UpdateOrder = 2208  # 订单状态变动通知(推送) """
        from futu.common.pb.Trd_UpdateOrder_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_UpdateOrder] = Response()

        """ Trd_GetOrderFillList = 2211  # 获取成交列表 """
        from futu.common.pb.Trd_GetOrderFillList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetOrderFillList] = Response()

        """ Trd_UpdateOrderFill = 2218  # 成交通知(推送) """
        from  futu.common.pb.Trd_UpdateOrderFill_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_UpdateOrderFill] = Response()

        """ Trd_GetHistoryOrderList = 2221  # 获取历史订单列表 """
        from futu.common.pb.Trd_GetHistoryOrderList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetHistoryOrderList] = Response()

        """ Trd_GetHistoryOrderFillList = 2222  # 获取历史成交列表 """
        from futu.common.pb.Trd_GetHistoryOrderFillList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetHistoryOrderFillList] = Response()

        """ Qot_Sub = 3001  # 订阅或者反订阅 """
        from futu.common.pb.Qot_Sub_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_Sub] = Response()

        """ Qot_RegQotPush = 3002  # 注册推送 """
        from futu.common.pb.Qot_RegQotPush_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_RegQotPush] = Response()

        """ Qot_GetSubInfo = 3003  # 获取订阅信息 """
        from futu.common.pb.Qot_GetSubInfo_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetSubInfo] = Response()

        """ Qot_GetBasicQot = 3004  # 获取股票基本行情 """
        from futu.common.pb.Qot_GetBasicQot_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetBasicQot] = Response()

        """ Qot_UpdateBasicQot = 3005  # 推送股票基本行情 """
        from futu.common.pb.Qot_UpdateBasicQot_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdateBasicQot] = Response()

        """ Qot_GetKL = 3006  # 获取K线 """
        from futu.common.pb.Qot_GetKL_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetKL] = Response()

        """ Qot_UpdateKL = 3007  # 推送K线 """
        from futu.common.pb.Qot_UpdateKL_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdateKL] = Response()

        """ Qot_GetRT = 3008  # 获取分时 """
        from futu.common.pb.Qot_GetRT_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetRT] = Response()

        """ Qot_UpdateRT = 3009  # 推送分时 """
        from futu.common.pb.Qot_UpdateRT_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdateRT] = Response()

        """ Qot_GetTicker = 3010  # 获取逐笔 """
        from futu.common.pb.Qot_GetTicker_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetTicker] = Response()

        """ Qot_UpdateTicker = 3011  # 推送逐笔 """
        from futu.common.pb.Qot_UpdateTicker_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdateTicker] = Response()

        """ Qot_GetOrderBook = 3012  # 获取买卖盘 """
        from futu.common.pb.Qot_GetOrderBook_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetOrderBook] = Response()

        """ Qot_UpdateOrderBook = 3013  # 推送买卖盘 """
        from futu.common.pb.Qot_UpdateOrderBook_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdateOrderBook] = Response()

        """ Qot_GetBroker = 3014  # 获取经纪队列 """
        from futu.common.pb.Qot_GetBroker_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetBroker] = Response()

        """ Qot_UpdateBroker = 3015  # 推送经纪队列 """
        from futu.common.pb.Qot_UpdateBroker_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdateBroker] = Response()

        """ Qot_UpdatePriceReminder = 3019  # 推送到价提醒 """
        from futu.common.pb.Qot_UpdatePriceReminder_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdatePriceReminder] = Response()


        """ Qot_GetTradeDate = 3200  # 获取市场交易日 """
        from futu.common.pb.Qot_GetTradeDate_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetTradeDate] = Response()

        """ Qot_GetSuspend = 3201  # 获取股票停牌信息 """
        from futu.common.pb.Qot_GetSuspend_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetSuspend] = Response()

        """ Qot_GetStaticInfo = 3202  # 获取股票列表 """
        from futu.common.pb.Qot_GetStaticInfo_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetStaticInfo] = Response()

        """ Qot_GetSecuritySnapshot = 3203  # 获取股票快照 """
        from futu.common.pb.Qot_GetSecuritySnapshot_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetSecuritySnapshot] = Response()

        """ Qot_GetPlateSet = 3204  # 获取板块集合下的板块 """
        from futu.common.pb.Qot_GetPlateSet_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetPlateSet] = Response()

        """ Qot_GetPlateSecurity = 3205  # 获取板块下的股票 """
        from futu.common.pb.Qot_GetPlateSecurity_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetPlateSecurity] = Response()

        """ Trd_GetMaxTrdQtys = 2111 查询最大买卖数量 """
        from futu.common.pb.Trd_GetMaxTrdQtys_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Trd_GetMaxTrdQtys] = Response()

        """ Qot_GetReference = 3206  获取正股相关股票，暂时只有窝轮"""
        from futu.common.pb.Qot_GetReference_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetReference] = Response()

        """ Qot_GetOwnerPlate = 3207 获取股票所属板块"""
        from futu.common.pb.Qot_GetOwnerPlate_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetOwnerPlate] = Response()

        """ Qot_GetOwnerPlate = 3208 获取高管持股变动"""
        from futu.common.pb.Qot_GetHoldingChangeList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetHoldingChangeList] = Response()

        from futu.common.pb.Qot_RequestHistoryKL_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_RequestHistoryKL] = Response()

        from futu.common.pb.Qot_GetOptionChain_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetOptionChain] = Response()

        """ Qot_GetOrderDetail = 3016 获取委托明细 """
        from futu.common.pb.Qot_GetOrderDetail_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetOrderDetail] = Response()

        """ Qot_UpdateOrderDetail = 3017 推送委托明细 """
        from futu.common.pb.Qot_UpdateOrderDetail_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_UpdateOrderDetail] = Response()

        """ Qot_GetWarrantData = 3210 获取涡轮 """
        from futu.common.pb.Qot_GetWarrant_pb2 import Response as GetWarrantPBResponse
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetWarrant] = GetWarrantPBResponse()

        """ Qot_GetOrderDetail = 3104 已使用过的额度 """
        from futu.common.pb.Qot_RequestHistoryKLQuota_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_RequestHistoryKLQuota] = Response()

        """获取除权信息"""
        from futu.common.pb.Qot_RequestRehab_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_RequestRehab] = Response()

        from futu.common.pb.Qot_GetCapitalDistribution_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetCapitalDistribution] = Response()

        from futu.common.pb.Qot_GetCapitalFlow_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetCapitalFlow] = Response()

        from futu.common.pb.Qot_GetUserSecurity_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetUserSecurity] = Response()

        from futu.common.pb.Qot_ModifyUserSecurity_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_ModifyUserSecurity] = Response()

        from futu.common.pb.Qot_StockFilter_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_StockFilter] = Response()

        from futu.common.pb.Qot_GetCodeChange_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetCodeChange] = Response()

        from futu.common.pb.Qot_GetIpoList_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetIpoList] = Response()
        
        from futu.common.pb.Qot_GetFutureInfo_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetFutureInfo] = Response()

        from futu.common.pb.Qot_RequestTradeDate_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_RequestTradeDate] = Response()

        from futu.common.pb.Qot_SetPriceReminder_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_SetPriceReminder] = Response()

        from futu.common.pb.Qot_GetPriceReminder_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetPriceReminder] = Response()

        from futu.common.pb.Qot_GetUserSecurityGroup_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetUserSecurityGroup] = Response()

        from futu.common.pb.Qot_GetMarketState_pb2 import Response
        ProtobufMap.created_protobuf_map[ProtoId.Qot_GetMarketState] = Response()

    def __getitem__(self, key):
        return ProtobufMap.created_protobuf_map[key] if key in ProtobufMap.created_protobuf_map else None


pb_map = ProtobufMap()

def binary2str(b, proto_id, proto_fmt_type):
    """
    Transfer binary to string
    :param b: binary content to be transformed to string
    :return: string
    """
    if proto_fmt_type == ProtoFMT.Json:
        return b.decode('utf-8')
    elif proto_fmt_type == ProtoFMT.Protobuf:
        rsp = pb_map[proto_id]
        if IS_PY2:
            rsp.ParseFromString(str(b))
        else:
            rsp.ParseFromString(b)
        return MessageToJson(rsp)
    else:
        raise Exception("binary2str: unknown proto format.")


def binary2pb(b, proto_id, proto_fmt_type):
    """
    Transfer binary to pb message
    :param b: binary content to be transformed to pb message
    :return: pb message
    """
    rsp = pb_map[proto_id]
    if rsp is None:
        return None
    if proto_fmt_type == ProtoFMT.Json:
        return json2pb(type(rsp), b.decode('utf-8'))
    elif proto_fmt_type == ProtoFMT.Protobuf:
        try:
            rsp = type(rsp)()
            # logger.debug((proto_id))
            if IS_PY2:
                rsp.ParseFromString(str(b))
            else:
                rsp.ParseFromString(b)
        except Exception as e:
            print(e)
        return rsp
    else:
        raise Exception("binary2str: unknown proto format.")


def pack_pb_req(pb_req, proto_id, conn_id, serial_no=0):
    proto_fmt = SysConfig.get_proto_fmt()
    serial_no = serial_no if serial_no else get_unique_id32()
    is_encrypt = FutuConnMng.is_conn_encrypt(conn_id)

    if proto_fmt == ProtoFMT.Json:
        req_json = MessageToJson(pb_req)
        ret, msg, req = _joint_head(proto_id, proto_fmt, len(req_json),
                          req_json.encode(), conn_id, serial_no, is_encrypt)
        return ret, msg, req

    elif proto_fmt == ProtoFMT.Protobuf:
        ret, msg, req = _joint_head(proto_id, proto_fmt, pb_req.ByteSize(), pb_req, conn_id, serial_no, is_encrypt)
        return ret, msg, req
    else:
        error_str = ERROR_STR_PREFIX + 'unknown protocol format, %d' % proto_fmt
        return RET_ERROR, error_str, None


def _joint_head(proto_id, proto_fmt_type, body_len, str_body, conn_id, serial_no, is_encrypt):

    # sha20 = b'00000000000000000000'
    reserve8 = b'\x00\x00\x00\x00\x00\x00\x00\x00'

    if proto_fmt_type == ProtoFMT.Protobuf:
        str_body = str_body.SerializeToString()

    if type(str_body) is not bytes:
        str_body = bytes_utf8(str_body)
    sha20 = hashlib.sha1(str_body).digest()

    # init connect 需要用rsa加密
    try:
        if proto_id == ProtoId.InitConnect:
            if SysConfig.INIT_RSA_FILE != '':
                str_body = RsaCrypt.encrypt(str_body)
                body_len = len(str_body)
        else:
            if is_encrypt:
                ret, msg, str_body = FutuConnMng.encrypt_conn_data(conn_id, str_body)
                body_len = len(str_body)
                if ret != RET_OK:
                    return ret, msg, str_body
    except Exception as e:
        return RET_ERROR, str(e), None

    fmt = "%s%ds" % (MESSAGE_HEAD_FMT, body_len)

    bin_head = struct.pack(fmt, b'F', b'T', proto_id, proto_fmt_type,
                           API_PROTO_VER, serial_no, body_len, sha20, reserve8, str_body)

    return RET_OK, "", bin_head


def parse_head(head_bytes):
    head_dict = {}
    head_dict['head_1'], head_dict['head_2'], head_dict['proto_id'], \
    head_dict['proto_fmt_type'], head_dict['proto_ver'], \
    head_dict['serial_no'], head_dict['body_len'], head_dict['sha20'], \
    head_dict['reserve8'], = struct.unpack(MESSAGE_HEAD_FMT, head_bytes)
    return head_dict


def parse_proto_info(head_bytes):
    unpacked = struct.unpack(MESSAGE_HEAD_FMT, head_bytes)
    return ProtoInfo(unpacked[2], unpacked[5])


def decrypt_rsp_body(rsp_body, head_dict, conn_id, is_encrypt):
    ret_code = RET_OK
    msg = ''
    sha20 = head_dict['sha20']
    proto_id = head_dict['proto_id']

    if is_encrypt:
        try:
            if proto_id == ProtoId.InitConnect:
                rsp_body = RsaCrypt.decrypt(rsp_body)
            else:
                ret_code, msg, decrypt_data = FutuConnMng.decrypt_conn_data(conn_id, rsp_body)
                rsp_body = decrypt_data

        except Exception as e:
            msg = sys.exc_info()[1]
            ret_code = RET_ERROR

    # check sha20
    if ret_code == RET_OK:
        sha20_check = hashlib.sha1(rsp_body).digest()
        if sha20_check != sha20:
            ret_code = RET_ERROR
            msg = "proto id:{} conn_id:{} check sha error!".format(proto_id, conn_id)

    return ret_code, msg, rsp_body


def make_from_namedtuple(t, **kwargs):
    """
    t是namedtuple，复制一份t，但其中部分字段更新为kwargs的值
    :param t:
    :param kwargs:
    :return:
    """
    d = t._asdict()
    d.update(kwargs)
    cls = type(t)
    return cls(**d)